﻿//---------------------------------------------------------------------
// <copyright file="MockStreamBasedJsonWriterFactoryWrapper.cs" company="Microsoft">
//      Copyright (C) Microsoft Corporation. All rights reserved. See License.txt in the project root for license information.
// </copyright>
//---------------------------------------------------------------------

#if NETCOREAPP3_1_OR_GREATER
using System;
using System.IO;
using System.Text;
using Microsoft.OData.Json;

namespace Microsoft.OData.Tests.Json
{
    /// <summary>
    /// This wraps an <see cref="IStreamBasedJsonWriterFactory"/> to
    /// intercept calls made to create a writer in order to allow
    /// testing code paths around JSON writer construction.
    /// </summary>
    internal sealed class MockStreamBasedJsonWriterFactoryWrapper : IStreamBasedJsonWriterFactory
    {
        private readonly IStreamBasedJsonWriterFactory innerFactory;

        public MockStreamBasedJsonWriterFactoryWrapper(IStreamBasedJsonWriterFactory wrappedFactory)
        {
            if (wrappedFactory == null)
            {
                throw new ArgumentNullException(nameof(wrappedFactory));
            }

            this.innerFactory = wrappedFactory;
        }

        public IJsonWriter CreateJsonWriter(Stream stream, bool isIeee754Compatible, Encoding encoding)
        {
            this.NumCalls++;
            this.Encoding = encoding;
            IJsonWriter writer = innerFactory.CreateJsonWriter(stream, isIeee754Compatible, encoding);
            this.CreatedWriter = writer;

            return writer;
        }

        public IJsonWriterAsync CreateAsynchronousJsonWriter(Stream stream, bool isIeee754Compatible, Encoding encoding)
        {
            this.NumCalls++;
            this.Encoding = encoding;
            IJsonWriterAsync writer = innerFactory.CreateAsynchronousJsonWriter(stream, isIeee754Compatible, encoding);
            this.CreatedAsyncWriter = writer;

            return writer;
        }

        /// <summary>
        /// The <see cref="IJsonWriter"/> that was last created by the wrapped <see cref="IStreamBasedJsonWriterFactory"/>.
        /// </summary>
        public IJsonWriter CreatedWriter { get; private set; }

        /// <summary>
        /// The <see cref="IJsonWriterAsync" that was last created by the wrapped <see cref="IStreamBasedJsonWriterFactory"/>.
        /// </summary>
        public IJsonWriterAsync CreatedAsyncWriter { get; private set; }

        /// <summary>
        /// The encoding used when creating the <see cref="IJsonWriter"/>.
        /// </summary>
        public Encoding Encoding { get; private set; }

        /// <summary>
        /// Number of times the <see cref="CreateJsonWriter(Stream, bool, Encoding)"/> method was called.
        /// </summary>
        public int NumCalls { get; private set; }
    }
}
#endif
