﻿//---------------------------------------------------------------------
// <copyright file="ClientAnnotationTests.cs" company="Microsoft">
//      Copyright (C) Microsoft Corporation. All rights reserved. See License.txt in the project root for license information.
// </copyright>
//---------------------------------------------------------------------

namespace Microsoft.OData.Client.TDDUnitTests.Tests.Annotation
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Linq;
    using Microsoft.OData.Client;
    using Xunit;

    public class ClientAnnotationTests
    {
        DefaultContainerPlus dsc = new DefaultContainerPlus(new Uri("http://odataService/"));
        IEnumerable<MergeOption> mergeOptions = Enum.GetValues(typeof(MergeOption)).Cast<MergeOption>();

        const string FeedPayloadWithInstanceAnnotationOnFeedandEntry = @"{
    ""@odata.context"":""http://odataService/$metadata#People"",
    ""@Microsoft.OData.SampleService.Models.TripPin.Gender"":""Male"",
    ""value"": [
        {
            ""@odata.type"": ""#Microsoft.OData.SampleService.Models.TripPin.Person"",
            ""@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
            [
                ""Bob1"",
                ""Cat1""
            ],
            ""UserName"":""BobCat"",
            ""FirstName"": ""Bob"",
            ""LastName"": ""Cat"",
            ""Location"": null,
            ""CompanyLocation"": null,
            ""Gender"":""Male"",
            ""Age"":""22"",
            ""Concurrency"":635548229917715070
        },
        {
            ""@odata.type"": ""#Microsoft.OData.SampleService.Models.TripPin.VipCustomer"",
            ""@Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2"":true,
            ""UserName"": ""JillJones"",
            ""FirstName"": ""Jill"",
            ""LastName"": ""Jones"",
            ""Location"": null,
            ""CompanyLocation"": null,
            ""Gender"":""Male"",
            ""Age"":""22"",
            ""Concurrency"":635548229917715070
        }
    ]
}";

        const string EntryPayloadWithInstanceAnnotationOnEntry = @"{
    ""@odata.context"":""http://odataService/$metadata#People/$entity"",
    ""@Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2"":false,
    ""UserName"":""aprilcline"",
    ""FirstName"":""April"",
    ""LastName"":""Cline"",
    ""Emails@Org.OData.Core.V1.Computed"":false,
    ""Emails@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
    [
        ""test""
    ],
    ""Emails"":[""Russell@abc.com""],
    ""Location"":
    {
        ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":""MyCity"",
        ""Address"":""1 Microsoft Way"",
        ""City"":
        {
            ""Name"":""Beijing"",
            ""CountryRegion"":""China""
        }
     },
    ""Gender"":""Female"",
    ""Concurrency@Org.OData.Core.V1.Computed"":""false"",
    ""Concurrency"":635549965782056080,
    ""Assets@Org.OData.Core.V1.Description"":""All assets"",
    ""Assets"":
    [
        {
            ""@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
            [
                ""PC1""
            ],
            ""Name"":""PC1"",
            ""Kind"":
            [
                ""PersonalAsset"",
                ""NormalAsset""
            ]
        },
        {
            ""@Org.OData.Core.V1.Description"": null,
            ""Name"":""PC2"",
            ""Kind@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
            [
                ""PersonalAsset""
            ],
            ""Kind"":
            [
                ""PersonalAsset"",
                ""NormalAsset""
            ]
        }
    ]
}";

        [Fact]
        public void TestGetAnnotationOnFeed()
        {
            foreach (var mergeOption in mergeOptions)
            {
                string response = FeedPayloadWithInstanceAnnotationOnFeedandEntry;
                TestAnnotation(response, () =>
                {
                    dsc.MergeOption = mergeOption;
                    var peopleQueryResponse1 = dsc.PeoplePlus.Execute();
                    var people1 = peopleQueryResponse1.ToList();

                    // Get Enum instance annotation on feed
                    PersonGenderPlus annotation;
                    bool result = dsc.TryGetAnnotation<PersonGenderPlus>(peopleQueryResponse1, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                    Assert.True(result);
                    Assert.Equal(PersonGenderPlus.MalePlus, annotation); ;

                    // Get collection instance annotation on one of the entities in the returned collection
                    ICollection<string> annotation2 = null;
                    result = dsc.TryGetAnnotation<ICollection<string>>(people1[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation2);
                    Assert.True(result);
                    Assert.NotNull(annotation2);
                    Assert.Equal("Bob1", annotation2.First());

                    // Set the response to a new value
                    response = response.Replace(
                        @"""@Microsoft.OData.SampleService.Models.TripPin.Gender"":""Male""",
                        @"""@Microsoft.OData.SampleService.Models.TripPin.Gender"":""Female""")
                        .Replace(@"""Bob1""", @"""Bob2""");
                    SetResponse(response);

                    // Query the entityset again
                    var peopleQueryResponse2 = dsc.PeoplePlus.Execute();
                    var people2 = peopleQueryResponse2.ToList();

                    // Verify the instance annotation for this feed has been changed
                    result = dsc.TryGetAnnotation<PersonGenderPlus>(peopleQueryResponse2, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                    Assert.True(result);
                    Assert.Equal(PersonGenderPlus.FemalePlus, annotation);

                    // Verify the instance annotation that is being tracked according to the merge option.
                    result = dsc.TryGetAnnotation<ICollection<string>>(people1[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation2);
                    Assert.True(result);
                    Assert.NotNull(annotation2);
                    if (mergeOption == MergeOption.OverwriteChanges || mergeOption == MergeOption.PreserveChanges)
                    {
                        Assert.Equal("Bob2", annotation2.First());
                    }
                    else
                    {
                        Assert.Equal("Bob1", annotation2.First());
                    }

                    result = dsc.TryGetAnnotation<ICollection<string>>(people2[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation2);
                    Assert.True(result);
                    Assert.NotNull(annotation2);
                    if (mergeOption == MergeOption.AppendOnly)
                    {
                        Assert.Equal("Bob1", annotation2.First());
                    }
                    else
                    {
                        Assert.Equal("Bob2", annotation2.First());
                    }
                });
            }
        }

        [Fact]
        public void TestGetAnnotationOnFeedWithFilter()
        {
            TestAnnotation(FeedPayloadWithInstanceAnnotationOnFeedandEntry, () =>
            {
                var peopleQueryResponse = ((DataServiceQuery<PersonPlus>)dsc.PeoplePlus.Where(p => p.FirstNamePlus == "Bob")).Execute();
                var people = peopleQueryResponse.ToList();

                ICollection<string> annotation2 = null;
                bool result = dsc.TryGetAnnotation<ICollection<string>>(people[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation2);
                Assert.True(result);
                Assert.NotNull(annotation2);
                Assert.Equal("Bob1", annotation2.First());

                // Verify the instance annotation for the colletion of entity
                PersonGenderPlus annotation;
                result = dsc.TryGetAnnotation<PersonGenderPlus>(peopleQueryResponse, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                Assert.True(result);
                Assert.Equal(PersonGenderPlus.MalePlus, annotation);
            });
        }

        [Fact]
        public void TestGetAnnotationOnFeedWithProjection()
        {
            var response = @"{
    ""@odata.context"":""http://odataService/$metadata#People(UserName,FirstName,Emails,Location,Concurrency)"",
    ""@Microsoft.OData.SampleService.Models.TripPin.Gender"":""FemalePlus"",
    ""value"": [
        {
            ""@Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2"":true,
            ""UserName"":""BobCat"",
            ""FirstName"": ""Bob"",
            ""Emails@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
            [
                ""test""
            ],
            ""Emails"":[""Russell@abc.com""],
            ""Location"":
            {
                ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":""MyCity"",
                ""Address"":""1 Microsoft Way"",
                ""City"":
                {
                    ""Name"":""Beijing"",
                    ""CountryRegion"":""China""
                }
             },
            ""Concurrency@Org.OData.Core.V1.Computed"":""false"",
            ""Concurrency"":635548229917715070
        },
        {
            ""@Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2"":false,
            ""UserName"": ""JillJones"",
            ""FirstName"": ""Jill"",
            ""Emails"":[""Russell@abc.com""],
            ""Location"":
            {
                ""Address"":""1 Microsoft Way"",
                ""City"":
                {
                    ""Name"":""Beijing"",
                    ""CountryRegion"":""China""
                }
             },
            ""Concurrency"":635548229917715070
        }
    ]
}";
            TestAnnotation(response, () =>
            {
                var peopleQueryResponse = ((DataServiceQuery<PersonPlus>)dsc.PeoplePlus
                    .Select(p =>
                        new PersonPlus
                        {
                            UserNamePlus = p.UserNamePlus,
                            FirstNamePlus = p.FirstNamePlus,
                            ConcurrencyPlus = p.ConcurrencyPlus,
                            EmailsPlus = p.EmailsPlus,
                            LocationPlus = p.LocationPlus,
                        }))
                    .Execute();
                var people = peopleQueryResponse.ToList();

                PersonGenderPlus annotation;

                // Verify the instance annotation for the collection of entity
                bool result = dsc.TryGetAnnotation(peopleQueryResponse, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                Assert.True(result);
                Assert.Equal(PersonGenderPlus.FemalePlus, annotation);

                // Verify the instance annotation for one of the entities in the returned collection.
                bool searchAnnotation;
                result = dsc.TryGetAnnotation((object)people[0], "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out searchAnnotation);
                Assert.True(result);
                Assert.True(searchAnnotation);

                // Verify the instance annotation for the property in one entity
                ICollection<string> seoAnnotation = null;
                result = dsc.TryGetAnnotation<Func<ICollection<string>>, ICollection<string>>(() => people[0].EmailsPlus, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out seoAnnotation);
                Assert.True(result);
                Assert.Equal("test", seoAnnotation.First());

                string cityNameAnnotation = null;
                result = dsc.TryGetAnnotation<Func<LocationPlus>, string>(() => people.First().LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out cityNameAnnotation);
                Assert.True(result);
                Assert.Equal("MyCity", cityNameAnnotation);

                bool computedAnnotation;
                result = dsc.TryGetAnnotation<Func<long>, bool>(() => people.First().ConcurrencyPlus, "Org.OData.Core.V1.Computed", out computedAnnotation);
                Assert.True(result);
                Assert.False(computedAnnotation);
            });
        }

        [Fact]
        public void TestGetAnnotationOnNavigationPropertyWithProjection()
        {
            var response = @"{
    ""@odata.context"":""http://odataService/$metadata#People(UserName,FirstName,Emails,Location,Concurrency,Partner)"",
    ""@Microsoft.OData.SampleService.Models.TripPin.Gender"":""FemalePlus"",
    ""value"": [
        {
            ""UserName"":""BobCat"",
            ""FirstName"": ""Bob"",
            ""Emails"":[""Russell@abc.com""],
            ""Location"":
            {
                ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":""MyCity"",
                ""Address"":""1 Microsoft Way"",
                ""City"":
                {
                    ""Name"":""Beijing"",
                    ""CountryRegion"":""China""
                }
             },
            ""Concurrency@Org.OData.Core.V1.Computed"":""false"",
            ""Concurrency"":635548229917715070,
            ""Partner"":
            {
                ""UserName"":""Marchcline"",
                ""FirstName"":""March"",
                ""LastName"":""Cline"",
                ""Emails"":[],
                ""Location"":null,
                ""Gender"":""Female"",
                ""Concurrency"":635549965782056081
            }
        },
        {
            ""UserName"": ""JillJones"",
            ""FirstName"": ""Jill"",
            ""Emails"":[""Russell@abc.com""],
            ""Location"":
            {
                ""Address"":""1 Microsoft Way"",
                ""City"":
                {
                    ""Name"":""Beijing"",
                    ""CountryRegion"":""China""
                }
             },
            ""Concurrency"":635548229917715070,
            ""Partner"":
            {
                ""@Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2"":true,
                ""UserName"":""Marchcline"",
                ""FirstName"":""March"",
                ""LastName"":""Cline"",
                ""Emails@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
                [
                    ""test""
                ],
                ""Emails"":[""Russell@abc.com""],
                ""Location"":
                {
                    ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":""MyCity"",
                    ""Address"":""1 Microsoft Way"",
                    ""City"":
                    {
                        ""Name"":""Beijing"",
                        ""CountryRegion"":""China""
                    }
                 },
                ""Gender"":""Female"",
                ""Concurrency@Org.OData.Core.V1.Computed"":""false"",
                ""Concurrency"":635548229917715070
            }
        }
    ]
}";
            TestAnnotation(response, () =>
            {
                var peopleQueryResponse = ((DataServiceQuery<PersonPlus>)dsc.PeoplePlus
                    .Select(p =>
                        new PersonPlus
                        {
                            UserNamePlus = p.UserNamePlus,
                            FirstNamePlus = p.FirstNamePlus,
                            ConcurrencyPlus = p.ConcurrencyPlus,
                            EmailsPlus = p.EmailsPlus,
                            LocationPlus = p.LocationPlus,
                            PartnerPlus = p.PartnerPlus
                        }))
                    .Execute();
                var people = peopleQueryResponse.ToList();

                PersonGenderPlus annotation;

                // Verify the instance annotation for the collection of entity
                bool result = dsc.TryGetAnnotation(peopleQueryResponse, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                Assert.True(result);
                Assert.Equal(PersonGenderPlus.FemalePlus, annotation);

                // Verify the instance annotation for the entity in the returned collection.
                bool searchAnnotation;
                result = dsc.TryGetAnnotation<Func<PersonPlus>, bool>(() => people[1].PartnerPlus, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out searchAnnotation);
                Assert.True(result);
                Assert.True(searchAnnotation);

                // Verify the instance annotation for the property in one entity
                ICollection<string> seoAnnotation = null;
                result = dsc.TryGetAnnotation<Func<ICollection<string>>, ICollection<string>>(() => people[1].PartnerPlus.EmailsPlus, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out seoAnnotation);
                Assert.True(result);
                Assert.Equal("test", seoAnnotation.First());

                string cityNameAnnotation = null;
                result = dsc.TryGetAnnotation<Func<LocationPlus>, string>(() => people[1].PartnerPlus.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out cityNameAnnotation);
                Assert.True(result);
                Assert.Equal("MyCity", cityNameAnnotation);

                bool computedAnnotation;
                result = dsc.TryGetAnnotation<Func<long>, bool>(() => people[1].PartnerPlus.ConcurrencyPlus, "Org.OData.Core.V1.Computed", out computedAnnotation);
                Assert.True(result);
                Assert.False(computedAnnotation);
            });
        }

        [Fact]
        public void TestGetAnnotationOnSingleton()
        {
            foreach (var mergeOption in mergeOptions)
            {
                string response = EntryPayloadWithInstanceAnnotationOnEntry;
                TestAnnotation(response, () =>
                {
                    dsc.MergeOption = mergeOption;
                    var me = dsc.MePlus.GetValue();

                    bool annotation;

                    // Get complex type instance annotation on entity returned by querying Singleton
                    bool result = dsc.TryGetAnnotation(me, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation);
                    Assert.True(result);
                    Assert.False(annotation);

                    // Set the response to a new value
                    response = response.Replace(
                        @"Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2"":false,",
                        @"Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2"":true,");
                    SetResponse(response);

                    // Query again
                    var me2 = dsc.MePlus.GetValue();

                    result = dsc.TryGetAnnotation(me2, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation);
                    Assert.True(result);
                    if (mergeOption == MergeOption.AppendOnly)
                    {
                        Assert.False(annotation);
                    }
                    else
                    {
                        Assert.True(annotation);
                    }

                    // TODO : #625
//                    // Set the response to a new value
//                    response = response.Replace(
//            @"""Searchable"":false,
//        ""UnsupportedExpressions"":""NOT,OR""",
//            @"""Searchable"":true,
//        ""UnsupportedExpressions"":""OR""");
//                    SetResponse(response);

//                    // Query again
//                    var me2 = dsc.MePlus.GetValue();

//                    // Get complex type instance annotation on entity according to the merge option
//                    annotation = null;
//                    result = dsc.TryGetAnnotation(me, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions", out annotation);
//                    Assert.True(result);
//                    if (mergeOption == MergeOption.OverwriteChanges || mergeOption == MergeOption.PreserveChanges)
//                    {
//                        Assert.Equal(SearchExpressionsPlus.ORPlus, annotation.UnsupportedExpressionsPlus);
//                    }
//                    else
//                    {
//                        Assert.Equal(SearchExpressionsPlus.NOTPlus | SearchExpressionsPlus.ORPlus, annotation.UnsupportedExpressionsPlus);
//                    }

//                    // Get the new annotation on this entity
//                    result = dsc.TryGetAnnotation(me2, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions", out annotation);
//                    Assert.True(result);
//                    if (mergeOption == MergeOption.AppendOnly)
//                    {
//                        Assert.Equal(SearchExpressionsPlus.NOTPlus | SearchExpressionsPlus.ORPlus, annotation.UnsupportedExpressionsPlus);
//                    }
//                    else
//                    {
//                        Assert.Equal(SearchExpressionsPlus.ORPlus, annotation.UnsupportedExpressionsPlus);
//                    }
                });
            }
        }

        [Fact]
        public void TestGetCollectionInstanceAnnotationOnODataEntryWhenPayloadisODataEntryButNotSingleton()
        {
            foreach (var mergeOption in mergeOptions)
            {
                string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People/$entity"",
    ""@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
    [
        ""Russell1"",
        ""Whyte1""
    ],
    ""UserName"":""russellwhyte"",
    ""FirstName"":""Russell"",
    ""LastName"":""Whyte"",
    ""Emails"":[""Russell@example.com"",""Russell@contoso.com""],
    ""Location"":
    {
        ""Address"":""1 Microsoft Way"",
        ""City"":
        {
            ""Name"":""Nanjing"",
            ""CountryRegion"":""China""
        }
     },
     ""Gender"":""Male"",
     ""Age"":""22"",
     ""Concurrency"":635548229917715070
 }";
                TestAnnotation(response, () =>
                {
                    dsc.MergeOption = mergeOption;
                    var personQuery = dsc.PeoplePlus.ByKey("russellwhyte");
                    var person = personQuery.GetValue();
                    ICollection<string> annotation = null;

                    // Get instance annotation on a single entity
                    bool result = dsc.TryGetAnnotation<ICollection<string>>(person, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                    Assert.True(result);
                    Assert.NotNull(annotation);
                    Assert.Equal("Russell1", annotation.ElementAt(0));

                    response = response.Replace("Russell1", "Russell2");
                    SetResponse(response);

                    var personQuery2 = dsc.PeoplePlus.ByKey("russellwhyte");
                    var person2 = personQuery2.GetValue();

                    // Get instance annotation on a single entity, should be changed according to the merge options
                    result = dsc.TryGetAnnotation<ICollection<string>>(person, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                    Assert.True(result);
                    Assert.NotNull(annotation);
                    if (mergeOption == MergeOption.OverwriteChanges || mergeOption == MergeOption.PreserveChanges)
                    {
                        Assert.Equal("Russell2", annotation.ElementAt(0));
                    }
                    else
                    {
                        Assert.Equal("Russell1", annotation.ElementAt(0));
                    }

                    result = dsc.TryGetAnnotation<ICollection<string>>(person2, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                    Assert.True(result);
                    Assert.NotNull(annotation);
                    if (mergeOption == MergeOption.AppendOnly)
                    {
                        Assert.Equal("Russell1", annotation.ElementAt(0));
                    }
                    else
                    {
                        Assert.Equal("Russell2", annotation.ElementAt(0));
                    }

                    response = response.Replace("Russell2", "Russell3");
                    SetResponse(response);

                    var personQuery3 = dsc.PeoplePlus.ByKey("russellwhyte").Select(p => new { UN = p.UserNamePlus, FN = p.FirstNamePlus, CU = p.ConcurrencyPlus });
                    var person3 = personQuery3.GetValue();

                    // Get instance annotation on a single entity, it is not effected by the entity stored on client side.
                    result = dsc.TryGetAnnotation<ICollection<string>>(person2, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                    Assert.True(result);
                    Assert.NotNull(annotation);
                    if (mergeOption == MergeOption.AppendOnly)
                    {
                        Assert.Equal("Russell1", annotation.ElementAt(0));
                    }
                    else
                    {
                        Assert.Equal("Russell2", annotation.ElementAt(0));
                    }

                    // Get instance annotation on the projected single entity.
                    result = dsc.TryGetAnnotation<ICollection<string>>(person3, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                    Assert.NotNull(annotation);
                    Assert.Equal("Russell3", annotation.ElementAt(0));
                });
            }
        }

        [Fact]
        public void TestGetAnnotationOnDerivedEntity()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People/$entity"",
    ""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.VipCustomer"",
    ""@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
    [
        ""Russell1"",
        ""Whyte1""
    ],
    ""UserName"":""russellwhyte"",
    ""FirstName"":""Russell"",
    ""LastName"":""Whyte"",
    ""Emails"":[""Russell@example.com"",""Russell@contoso.com""],
    ""Location"":
    {
        ""Address"":""1 Microsoft Way"",
        ""City"":
        {
            ""Name"":""Nanjing"",
            ""CountryRegion"":""China""
        }
     },
     ""Gender"":""Male"",
     ""Age"":""22"",
     ""Concurrency"":635548229917715070,
     ""VipNumber"":1
 }";
            TestAnnotation(response, () =>
            {
                var person = dsc.PeoplePlus.ByKey("russellwhyte").GetValue();
                ICollection<string> annotation = null;

                // Get instance annotation on a single entity
                bool result = dsc.TryGetAnnotation<ICollection<string>>(person, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                Assert.True(result);
                Assert.NotNull(annotation);
                Assert.Equal("Russell1", annotation.ElementAt(0));
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnEntityWhichIsReturnedByFunction()
        {
            TestAnnotation(EntryPayloadWithInstanceAnnotationOnEntry, () =>
            {
                var partnerQuery = dsc.PeoplePlus.ByKey("russellwhyte").GetPartnerPlus();
                var partner = partnerQuery.GetValue();

                bool annotation;
                var result = dsc.TryGetAnnotation(partner, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation);
                Assert.True(result);
                Assert.False(annotation);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnCollectionOfEntityWhichIsReturnedByFunction()
        {
            TestAnnotation(FeedPayloadWithInstanceAnnotationOnFeedandEntry, () =>
            {
                var friendsQueryResponse = dsc.PeoplePlus.ByKey("russellwhyte").GetFriendsPlus().Execute();
                var friends = friendsQueryResponse.ToList();

                // Get complex type instance annotation on returned feed
                PersonGenderPlus annotation;
                bool result = dsc.TryGetAnnotation(friendsQueryResponse, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                Assert.True(result);
                Assert.Equal(PersonGenderPlus.MalePlus, annotation);

                // Get collection type instance annotation on one of the returned items
                ICollection<string> annotation1 = null;
                result = dsc.TryGetAnnotation(friends[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation1);
                Assert.True(result);
                Assert.NotNull(annotation1);
                Assert.Equal("Bob1", annotation1.ElementAt(0));

                // Get collection type instance annotation on one of the returned derived items
                bool annotation2;
                result = dsc.TryGetAnnotation(friends[1], "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation2);
                Assert.True(result);
                Assert.True(annotation2);
            });
        }

        const string ComplexValuePayloadWithInstanceAnnotationOnComplexValue = @"{
    ""@odata.context"":""http://odataService/$metadata#Microsoft.OData.SampleService.Models.TripPin.Location"",
    ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":""MyCity"",
    ""Address"":""1 Microsoft Way"",
    ""City"":
    {
        ""Name"":""Nanjing"",
        ""CountryRegion"":""China""
    }
}";
        [Fact]
        public void TestGetInstanceAnnotationOnODataComplexValueWhichIsReturnedByFunction()
        {
            TestAnnotation(ComplexValuePayloadWithInstanceAnnotationOnComplexValue, () =>
            {
                var getLocationQuery = dsc.PeoplePlus.ByKey("Jason").GetLocationPlus();
                var location = getLocationQuery.GetValue();

                string annotation = null;
                bool result = dsc.TryGetAnnotation<string>(location, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Equal("MyCity", annotation);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnEntityWhichIsReturnedByAction()
        {
            TestAnnotation(EntryPayloadWithInstanceAnnotationOnEntry, () =>
            {
                var PartnerAction = dsc.PeoplePlus.ByKey("russellwhyte").SetPartnerPlus();
                var partner = PartnerAction.GetValue();

                // Get complex type instance annotation on returned entity
                bool annotation;
                var result = dsc.TryGetAnnotation(partner, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation);
                Assert.True(result);
                Assert.False(annotation);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnCollectionOfEntityWhichIsReturnedByAction()
        {
            TestAnnotation(FeedPayloadWithInstanceAnnotationOnFeedandEntry, () =>
            {
                var friendsQuery = dsc.PeoplePlus.ByKey("russellwhyte").SetFriendsPlus();
                var friendsResponse = friendsQuery.Execute();
                var friends = friendsResponse.ToList();

                // Get complex type instance annotation on returned feed
                PersonGenderPlus annotation;
                bool result = dsc.TryGetAnnotation(friendsResponse, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                Assert.True(result);
                Assert.Equal(PersonGenderPlus.MalePlus, annotation);

                // Get collection type instance annotation on one of the returned items
                ICollection<string> annotation1 = null;
                result = dsc.TryGetAnnotation(friends[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation1);
                Assert.True(result);
                Assert.NotNull(annotation1);
                Assert.Equal("Bob1", annotation1.ElementAt(0));

                // Get collection type instance annotation on one of the returned derived items
                bool annotation2;
                result = dsc.TryGetAnnotation(friends[1], "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation2);
                Assert.True(result);
                Assert.True(annotation2);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnODataComplexValueWhichIsReturnedByAction()
        {
            TestAnnotation(ComplexValuePayloadWithInstanceAnnotationOnComplexValue, () =>
            {
                var getLocationQuery = dsc.PeoplePlus.ByKey("Jason").SetLocationPlus();
                var location = getLocationQuery.GetValue();

                string annotation = null;
                bool result = dsc.TryGetAnnotation<string>(location, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Equal("MyCity", annotation);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnEntityWhichIsReturnedByFunctionImport()
        {
            TestAnnotation(EntryPayloadWithInstanceAnnotationOnEntry, () =>
            {
                var getPersonQuery = dsc.GetPersonPlus("Russell");
                var person = getPersonQuery.GetValue();

                // Get complex type instance annotation on returned entity
                bool annotation;
                var result = dsc.TryGetAnnotation(person, "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation);
                Assert.True(result);
                Assert.False(annotation);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnCollectionOfEntityWhichIsReturnedByFunctionImport()
        {
            TestAnnotation(FeedPayloadWithInstanceAnnotationOnFeedandEntry, () =>
            {
                var getFriendsQueryResponse = dsc.GetFriendsPlus("Russell").Execute() as QueryOperationResponse<PersonPlus>;
                var friends = getFriendsQueryResponse.ToList();

                // Get complex type instance annotation on returned feed
                PersonGenderPlus annotation;
                bool result = dsc.TryGetAnnotation(getFriendsQueryResponse, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                Assert.True(result);
                Assert.Equal(PersonGenderPlus.MalePlus, annotation);

                // Get collection type instance annotation on one of the returned items
                ICollection<string> annotation1 = null;
                result = dsc.TryGetAnnotation(friends[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation1);
                Assert.True(result);
                Assert.NotNull(annotation1);
                Assert.Equal("Bob1", annotation1.ElementAt(0));

                // Get collection type instance annotation on one of the returned derived items
                bool annotation2;
                result = dsc.TryGetAnnotation(friends[1], "Microsoft.OData.SampleService.Models.TripPin.SearchRestrictions2", out annotation2);
                Assert.True(result);
                Assert.True(annotation2);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnComplexTypePropertyInAnEntity()
        {
            foreach (var mergeOption in mergeOptions)
            {
                string response = EntryPayloadWithInstanceAnnotationOnEntry;
                TestAnnotation(response, () =>
                {
                    dsc.MergeOption = mergeOption;
                    var personQuery = dsc.PeoplePlus.ByKey("russellwhyte");
                    var person1 = personQuery.GetValue();

                    // Get instance annotation on ComplexProperty
                    string annotation = null;
                    bool result = dsc.TryGetAnnotation<Func<LocationPlus>, string>(() => person1.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                    Assert.True(result);
                    Assert.Equal("MyCity", annotation);

                    response = response.Replace("MyCity", "MyCity1");
                    SetResponse(response);

                    var person2 = personQuery.GetValue();

                    string annotationOnComplexInstance = null;
                    // Get instance annotation on ComplexProperty of the old entity by using propertyAccess
                    result = dsc.TryGetAnnotation<Func<LocationPlus>, string>(() => person1.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                    Assert.True(result);

                    // Get instance annotation on ComplexProperty of the old entity by using property instance
                    result = dsc.TryGetAnnotation(person1.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotationOnComplexInstance);
                    Assert.True(result);
                    if (mergeOption == MergeOption.OverwriteChanges || mergeOption == MergeOption.PreserveChanges)
                    {
                        Assert.Equal("MyCity1", annotation);
                        Assert.Equal("MyCity1", annotationOnComplexInstance);
                    }
                    else
                    {
                        Assert.Equal("MyCity", annotation);
                        Assert.Equal("MyCity", annotationOnComplexInstance);
                    }

                    // Get instance annotation on ComplexProperty of the new entity by using propertyAccess
                    result = dsc.TryGetAnnotation<Func<LocationPlus>, string>(() => person2.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                    Assert.True(result);

                    // Get instance annotation on ComplexProperty of the new entity by using property instance
                    result = dsc.TryGetAnnotation(person2.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotationOnComplexInstance);
                    Assert.True(result);
                    if (mergeOption == MergeOption.AppendOnly)
                    {
                        Assert.Equal("MyCity", annotation);
                        Assert.Equal("MyCity", annotationOnComplexInstance);
                    }
                    else
                    {
                        Assert.Equal("MyCity1", annotation);
                        Assert.Equal("MyCity1", annotationOnComplexInstance);
                    }
                });
            }
        }

        [Fact]
        public void TestGetAnnotationOnPrimitivePropertyInAnEntity()
        {
            TestAnnotation(EntryPayloadWithInstanceAnnotationOnEntry, () =>
            {
                var person = dsc.PeoplePlus.ByKey("russellwhyte").GetValue();

                bool annotation;
                // Get instance annotation on primitive property
                bool result = dsc.TryGetAnnotation<Func<long>, bool>(() => person.ConcurrencyPlus, "Org.OData.Core.V1.Computed", out annotation);
                Assert.True(result);
                Assert.False(annotation);
            });
        }

        [Fact]
        public void TestGetAnnotationOnCollectionPropertyInAnEntity()
        {
            TestAnnotation(EntryPayloadWithInstanceAnnotationOnEntry, () =>
            {
                var person = dsc.PeoplePlus.ByKey("russellwhyte").GetValue();

                // Get instance annotation on collection property
                ICollection<string> annotation;
                bool result = dsc.TryGetAnnotation<Func<ObservableCollection<string>>, ICollection<string>>(() => person.EmailsPlus, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                Assert.True(result);
                Assert.NotNull(annotation);
                Assert.Equal("test", annotation.FirstOrDefault());

                // Get instance annotation on the same collection property
                bool annotation1;
                result = dsc.TryGetAnnotation<Func<ObservableCollection<string>>, bool>(() => person.EmailsPlus, "Org.OData.Core.V1.Computed", out annotation1);
                Assert.True(result);
                Assert.False(annotation1);
            });
        }

        // github: https://github.com/OData/odata.net/issues/879: Need to support instance annotations on feed or nestedResourceInfo.
        //  [Fact]
        internal void TestGetAnnotationOnCollectionOfComplexTypePropertyInAnEntity()
        {
            TestAnnotation(EntryPayloadWithInstanceAnnotationOnEntry, () =>
            {
                var person = dsc.PeoplePlus.ByKey("russellwhyte").GetValue();

                // Get instance annotation on collection property
                string annotation;
                bool result = dsc.TryGetAnnotation<Func<ObservableCollection<AssetPlus>>, string>(() => person.AssetsPlus, "Org.OData.Core.V1.Description", out annotation);
                Assert.True(result);
                Assert.Equal("All assets", annotation);

                ICollection<string> annotation1;
                // Get instance annotation on one item
                result = dsc.TryGetAnnotation(person.AssetsPlus[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation1);
                Assert.True(result);
                Assert.NotNull(annotation1);
                Assert.Equal("PC1", annotation1.FirstOrDefault());

                // Get null instance annotation on one item
                result = dsc.TryGetAnnotation(person.AssetsPlus[1], "Org.OData.Core.V1.Description", out annotation);
                Assert.True(result);
                Assert.Null(annotation);

                // Get instance annotation on the collection in the complex property
                result = dsc.TryGetAnnotation<Func<ObservableCollection<AssetTypePlus>>, ICollection<string>>(
                    () => person.AssetsPlus[1].KindPlus, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms",
                    out annotation1);
                Assert.True(result);
                Assert.NotNull(annotation1);
                Assert.Equal("PersonalAsset", annotation1.FirstOrDefault());
            });
        }

        [Fact]
        public void TestGetAnnotationOnPropertyInDerivedEntity()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People/$entity"",
    ""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.VipCustomer"",
    ""@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
    [
        ""Russell1"",
        ""Whyte1""
    ],
    ""UserName"":""russellwhyte"",
    ""FirstName"":""Russell"",
    ""LastName"":""Whyte"",
    ""Emails"":[""Russell@example.com"",""Russell@contoso.com""],
    ""Location"":
    {
        ""@Microsoft.OData.SampleService.Models.TripPin.CityName"" : ""Test1"",
        ""Address"":""1 Microsoft Way"",
        ""City"":
        {
            ""Name"":""Nanjing"",
            ""CountryRegion"":""China""
        }
     },
    ""CompanyLocation"":
    {
        ""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.CompanyLocation"",
        ""@Microsoft.OData.SampleService.Models.TripPin.CityName"" : ""Test2"",
        ""Address"":""1 Microsoft Way"",
        ""City"":
        {
            ""Name"":""Nanjing"",
            ""CountryRegion"":""China""
        }
     },
     ""Gender"":""Male"",
     ""Age"":""22"",
     ""Concurrency"":635548229917715070,
     ""VipNumber"":1
 }";
            TestAnnotation(response, () =>
            {
                var person = dsc.PeoplePlus.ByKey("russellwhyte").GetValue();
                string annotation = null;

                // Get instance annotation on a single entity
                bool result = dsc.TryGetAnnotation<Func<LocationPlus>, string>(() => person.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Equal("Test1", annotation);

                result = dsc.TryGetAnnotation<Func<LocationPlus>, string>(() => person.CompanyLocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Equal("Test2", annotation);
            });
        }
        
        [Fact]
        public void TestGetInstanceAnnotationsOnEntityAndPropertiesWithSameName()
        {
            string response = @"{
	""@odata.context"":""http://odataService/$metadata#People('Adrian')/Location"",
	""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.CompanyLocation"",
	""@Org.OData.Core.V1.Description"":""My Company"",
	""CompanyName@Org.OData.Core.V1.Description"":""My Company Name"",
	""CompanyName"":""HMC"",
	""Address@Org.OData.Core.V1.Description"":""My Company Address"",
	""Address"":""9 Seed Street"",
	""City"":
	{
	""Name"":""Crest"",
	""CountryRegion"":""France""
	}
	}";

            TestAnnotation(response, () =>
            {
                var location = dsc.PeoplePlus.ByKey("Adrian").Select(p => p.LocationPlus).GetValue() as CompanyLocationPlus;

                string annotation = null;
                bool result = false;

                result = dsc.TryGetAnnotation<string>(location, "Org.OData.Core.V1.Description", out annotation);
                Assert.True(result);
                Assert.Equal("My Company", annotation);

                result = dsc.TryGetAnnotation<Func<string>, string>(() => location.CompanyNamePlus, "Org.OData.Core.V1.Description", out annotation);
                Assert.True(result);
                Assert.Equal("My Company Name", annotation);

                result = dsc.TryGetAnnotation<Func<string>, string>(() => location.AddressPlus, "Org.OData.Core.V1.Description", out annotation);
                Assert.True(result);
                Assert.Equal("My Company Address", annotation);
            });
        }

        [Fact]
        public void TesDuplicateAnnotationsOnEntityThrows()
        {
            string response = @"{
	""@odata.context"":""http://odataService/$metadata#People('Adrian')/Location"",
	""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.CompanyLocation"",
	""@Org.OData.Core.V1.Description"":""My Company 1"",
	""@Org.OData.Core.V1.Description"":""My Company 2"",
	""CompanyName"":""HMC"",
	""Address"":""9 Seed Street"",
	""City"":
	{
	""Name"":""Crest"",
	""CountryRegion"":""France""
	}
	}";

            TestAnnotation(response, () =>
            {
                Exception exception = null;

                try
                {
                    var location = dsc.PeoplePlus.ByKey("Adrian").Select(p => p.LocationPlus).GetValue() as CompanyLocationPlus;
                }
                catch (ArgumentException ex)
                {
                    Assert.Equal("An item with the same key has already been added.", ex.Message);
                    exception = ex;
                }

                Assert.NotNull(exception);
            });
        }

        [Fact]
        public void TestDuplicateAnnotationsOnPropertyThrows()
        {
            string response = @"{
	""@odata.context"":""http://odataService/$metadata#People('Adrian')/Location"",
	""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.CompanyLocation"",
	""CompanyName@Org.OData.Core.V1.Description"":""My Company Name 1"",
	""CompanyName@Org.OData.Core.V1.Description"":""My Company Name 2"",
	""CompanyName"":""HMC"",
	""Address"":""9 Seed Street"",
	""City"":
	{
	""Name"":""Crest"",
	""CountryRegion"":""France""
	}
	}";

            TestAnnotation(response, () =>
            {
                Exception exception = null;

                try
                {
                    var location = dsc.PeoplePlus.ByKey("Adrian").Select(p => p.LocationPlus).GetValue() as CompanyLocationPlus;
                }
                catch (ArgumentException ex)
                {
                    Assert.Equal("An item with the same key has already been added.", ex.Message);
                    exception = ex;
                }

                Assert.NotNull(exception);
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnSingleNavigationProperty()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#Me"",
    ""UserName"":""aprilcline"",
    ""FirstName"":""April"",
    ""LastName"":""Cline"",
    ""Emails"":[],
    ""Location"":null,
    ""Gender"":""Male"",
    ""Concurrency"":635549965782056080,
    ""Partner"":
    {
        ""@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
        [
            ""Bob1"",
            ""Cat1""
        ],
        ""UserName"":""Marchcline"",
        ""FirstName"":""March"",
        ""LastName"":""Cline"",
        ""Emails"":[],
        ""Location"":null,
        ""Gender"":""Female"",
        ""Concurrency"":635549965782056081
    }
}";
            TestAnnotation(response, () =>
            {
                var person = dsc.PeoplePlus.ByKey("russellwhyte").Expand(p => p.PartnerPlus).GetValue();
                ICollection<string> annotation = null;
                bool result = dsc.TryGetAnnotation<Func<PersonPlus>, ICollection<string>>(() => person.PartnerPlus, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                Assert.True(result);
                Assert.NotNull(annotation);
                Assert.Equal("Bob1", annotation.ElementAt(0));
                Assert.Equal("Cat1", annotation.ElementAt(1));

                annotation = null;
                result = dsc.TryGetAnnotation(person.PartnerPlus, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                Assert.True(result);
                Assert.NotNull(annotation);
                Assert.Equal("Bob1", annotation.ElementAt(0));
                Assert.Equal("Cat1", annotation.ElementAt(1));
            });
        }

        [Fact]
        public void TestGetInstanceAnnotationOnODataComplexValueWhenPayloadisODataComplexValue()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People('Jason')/Location"",
    ""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.Location"",
    ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":""MyCity"",
    ""Address"":""1 Microsoft Way"",
    ""City"":
    {
        ""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.City"",
        ""Name"":""Nanjing"",
        ""CountryRegion"":""China""
    }
}";
            TestAnnotation(response, () =>
            {
                var location = dsc.PeoplePlus.ByKey("Jason").Select(p => p.LocationPlus).GetValue();
                string annotation = null;
                bool result = dsc.TryGetAnnotation<string>(location, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Equal("MyCity", annotation);
            });
        }

        [Fact]
        public void TestGetStringInstanceAnnotationOnODataComplexValueWhenPayloadisODataComplexValueofDerivedType()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People('Jason')/Location"",
    ""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.CompanyLocation"",
    ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":""MyCity"",
    ""CompanyName@Org.OData.Core.V1.Description"":""My Company"",
    ""CompanyName"":""MATL"",
    ""Address"":""1 Microsoft Way"",
    ""City"":
    {
        ""Name"":""Nanjing"",
        ""CountryRegion"":""China""
    }
}";
            TestAnnotation(response, () =>
            {
                var location = dsc.PeoplePlus.ByKey("Jason").Select(p => p.LocationPlus).GetValue() as CompanyLocationPlus;
                string annotation = null;
                bool result = dsc.TryGetAnnotation<string>(location, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Equal("MyCity", annotation);

                // Verify the instance annotaiton on property in derived type
                result = dsc.TryGetAnnotation<Func<string>, string>(() => location.CompanyNamePlus, "Org.OData.Core.V1.Description", out annotation);
                Assert.True(result);
                Assert.Equal("My Company", annotation);
            });
        }

        [Fact]
        public void TestGetNullableIntegerInstanceAnnotationOnODataEntry()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People/$entity"",
    ""@Microsoft.OData.SampleService.Models.TripPin.Age"":null,
    ""UserName"":""aprilcline"",
    ""FirstName"":""April"",
    ""LastName"":""Cline"",
    ""Emails"":[""aprilcline@abc.com""],
    ""Location"":
    {
        ""@Microsoft.OData.SampleService.Models.TripPin.CityName"":null,
        ""Address"":""Smoke-bag slanting street"",
        ""City"":
        {
            ""Name"":""Beijing"",
            ""CountryRegion"":""China""
        }
     },
    ""Gender"":""Female"",
    ""Concurrency@Org.OData.Core.V1.Computed"":null,
    ""Concurrency"":635549965782056080
}";
            TestAnnotation(response, () =>
            {
                var person = dsc.PeoplePlus.ByKey("aprilcline").GetValue();
                string annotation = null;
                bool result = dsc.TryGetAnnotation<string>(person.LocationPlus, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Null(annotation);

                Nullable<int> ageAnnotation = null;
                result = dsc.TryGetAnnotation<Nullable<int>>(person, "Microsoft.OData.SampleService.Models.TripPin.Age", out ageAnnotation);
                Assert.True(result);
                Assert.Null(ageAnnotation);
            });
        }

        [Fact]
        public void TestGetCollectionInstanceAnnotationWithOneNullItemOnEntity()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People/$entity"",
    ""@Microsoft.OData.SampleService.Models.TripPin.SeoTerms"":
    [
        ""Russell1"",
        null,
        ""Whyte1""
    ],
    ""UserName"":""russellwhyte"",
    ""FirstName"":""Russell"",
    ""LastName"":""Whyte"",
    ""Emails"":[""Russell@example.com"",""Russell@contoso.com""],
    ""Location"": null,
     ""Gender"":""Male"",
     ""Age"":""22"",
     ""Concurrency"":635548229917715070
 }";
            TestAnnotation(response, () =>
            {
                var personQuery = dsc.PeoplePlus.ByKey("russellwhyte");
                var person = personQuery.GetValue();
                ICollection<string> annotation = null;

                bool result = dsc.TryGetAnnotation<ICollection<string>>(person, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);

                Assert.True(result);
                Assert.NotNull(annotation);
                Assert.Equal(3, annotation.Count);
                Assert.Equal("Russell1", annotation.ElementAt(0));
                Assert.Null(annotation.ElementAt(1));
                Assert.Equal("Whyte1", annotation.ElementAt(2));
            });
        }

        // github: https://github.com/OData/odata.net/issues/879: Need to support instance annotations on feed or nestedResourceInfo.
        //  [Fact]
        internal void TestGetDerivedInstanceWhileTheTermIsInBaseTypeAnnotationOnODataEntry()
        {
            string response = @"{
    ""@odata.context"":""http://odataService/$metadata#People/$entity"",
    ""@Microsoft.OData.SampleService.Models.TripPin.UnsupportedExpressions"":
    {
        ""@odata.type"":""#Microsoft.OData.SampleService.Models.TripPin.SearchRestrictionsType"",
        ""Searchable"":false,
        ""UnsupportedExpressions"":""NOT,OR""
    },
    ""UserName"":""aprilcline"",
    ""FirstName"":""April"",
    ""LastName"":""Cline"",
    ""Emails"":[""Russell@abc.com""],
    ""Location"":null,
    ""Gender"":""Female"",
    ""Concurrency"":635549965782056080
}";
            string unsupportExpressionTerm = "Microsoft.OData.SampleService.Models.TripPin.UnsupportedExpressions";
            TestAnnotation(response, () =>
            {
                dsc.MergeOption = MergeOption.OverwriteChanges;
                var person = dsc.PeoplePlus.ByKey("Jason").GetValue();
                SearchRestrictionsTypePlus annotation = null;
                bool result = dsc.TryGetAnnotation(person, unsupportExpressionTerm, out annotation);
                Assert.True(result);
                Assert.False(annotation.SearchablePlus.Value);
                Assert.Equal(SearchExpressionsPlus.NOTPlus | SearchExpressionsPlus.ORPlus, annotation.UnsupportedExpressionsPlus);

                SearchRestrictionsTypePlus unsupportExpressionAnnotation = null;
                result = dsc.TryGetAnnotation(person, unsupportExpressionTerm, out unsupportExpressionAnnotation);
                Assert.True(result);
                var typeCastedAnnotation = unsupportExpressionAnnotation as SearchRestrictionsTypePlus;
                Assert.NotNull(typeCastedAnnotation);
                Assert.False(typeCastedAnnotation.SearchablePlus.Value);

                SetResponse(response.Replace(@"""Searchable"":false,", @"""Searchable"":null,"));

                person = dsc.PeoplePlus.ByKey("Jason").GetValue();
                result = dsc.TryGetAnnotation(person, unsupportExpressionTerm, out annotation);
                Assert.True(result);
                Assert.Null(annotation.SearchablePlus);
                Assert.Equal(SearchExpressionsPlus.NOTPlus | SearchExpressionsPlus.ORPlus, annotation.UnsupportedExpressionsPlus);

                result = dsc.TryGetAnnotation(person, unsupportExpressionTerm, out unsupportExpressionAnnotation);
                Assert.True(result);
                typeCastedAnnotation = unsupportExpressionAnnotation as SearchRestrictionsTypePlus;
                Assert.NotNull(typeCastedAnnotation);
                Assert.Null(typeCastedAnnotation.SearchablePlus);
            });
        }

        [Fact]
        public void TestDisableInstanceAnnotationMaterializationOnODataFeed()
        {
            string response = FeedPayloadWithInstanceAnnotationOnFeedandEntry;
            TestAnnotation(response, () =>
            {
                dsc.DisableInstanceAnnotationMaterialization = true;
                var peopleQueryResponse = dsc.PeoplePlus.Execute();
                var people = peopleQueryResponse.ToList();

                // Get Enum annotation on feed, no instance annotation. 
                PersonGenderPlus annotation;
                bool result = dsc.TryGetAnnotation<PersonGenderPlus>(peopleQueryResponse, "Microsoft.OData.SampleService.Models.TripPin.Gender", out annotation);
                Assert.False(result);
                Assert.Equal(PersonGenderPlus.UnknownPlus, annotation);

                // Get annotaion on entry, no instance annotation, return metadata annotation.
                ICollection<string> annotation2 = null;
                result = dsc.TryGetAnnotation<ICollection<string>>(people[0], "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation2);
                Assert.True(result);
                Assert.Equal("Bob", annotation2.First());
                Assert.Equal("BobCat", annotation2.ElementAt(1));
            });
        }

        [Fact]
        public void TestDisableInstanceAnnotationMaterializationOnOnODataComplexValue()
        {
            TestAnnotation(ComplexValuePayloadWithInstanceAnnotationOnComplexValue, () =>
            {
                dsc.DisableInstanceAnnotationMaterialization = true;
                var getLocationQuery = dsc.PeoplePlus.ByKey("Jason").SetLocationPlus();
                var location = getLocationQuery.GetValue();

                string annotation = null;
                bool result = dsc.TryGetAnnotation<string>(location, "Microsoft.OData.SampleService.Models.TripPin.CityName", out annotation);
                Assert.True(result);
                Assert.Equal("Nanjing", annotation);
            });
        }

        [Fact]
        public void TestDisableInstanceAnnotationMaterializationOnCollectionPropertyInAnEntity()
        {
            TestAnnotation(EntryPayloadWithInstanceAnnotationOnEntry, () =>
            {
                dsc.DisableInstanceAnnotationMaterialization = true;
                var person = dsc.PeoplePlus.ByKey("russellwhyte").GetValue();

                // Get instance annotation on collection property
                ICollection<string> annotation;
                bool result = dsc.TryGetAnnotation<Func<ObservableCollection<string>>, ICollection<string>>(() => person.EmailsPlus, "Microsoft.OData.SampleService.Models.TripPin.SeoTerms", out annotation);
                Assert.False(result);
                Assert.Null(annotation);

                // Get instance annotation on the same collection property
                bool annotation1;
                result = dsc.TryGetAnnotation<Func<ObservableCollection<string>>, bool>(() => person.EmailsPlus, "Org.OData.Core.V1.Computed", out annotation1);
                Assert.False(result);
                Assert.False(annotation1);
            });
        }

        [Fact]
        public void TestInstanceAnnotationsShouldBeEmptyIfNoInstanceAnnotationReturned()
        {
            const string FeedPayloadWithoutInstanceAnnotation = @"{
    ""@odata.context"":""http://odataService/$metadata#People"",
    ""value"": [
        {
            ""@odata.type"": ""#Microsoft.OData.SampleService.Models.TripPin.Person"",
            ""UserName"":""BobCat"",
            ""FirstName"": ""Bob"",
            ""LastName"": ""Cat"",
            ""Location"": null,
            ""CompanyLocation"": null,
            ""Gender"":""Male"",
            ""Age"":""22"",
            ""Concurrency"":635548229917715070
        },
        {
            ""@odata.type"": ""#Microsoft.OData.SampleService.Models.TripPin.VipCustomer"",
            ""UserName"": ""JillJones"",
            ""FirstName"": ""Jill"",
            ""LastName"": ""Jones"",
            ""Location"": null,
            ""CompanyLocation"": null,
            ""Gender"":""Male"",
            ""Age"":""22"",
            ""Concurrency"":635548229917715070
        }
    ]
}";

            TestAnnotation(FeedPayloadWithoutInstanceAnnotation, () =>
            {
                var getFriendsQueryResponse = dsc.GetFriendsPlus("Russell").Execute() as QueryOperationResponse<PersonPlus>;
                var friends = getFriendsQueryResponse.ToList();

                Assert.Equal(0, dsc.InstanceAnnotations.Count);
            });
        }

        private void TestAnnotation(string response, Action testAction)
        {
            dsc.DisableInstanceAnnotationMaterialization = false;
            SetResponse(response);

            dsc.SendingRequest2 += (sender, args) =>
            {
                args.RequestMessage.SetHeader("Prefer", "odata.include-annotations=\"*\"");
            };
            testAction();
        }

        private void SetResponse(string response)
        {
            dsc.Configurations.RequestPipeline.OnMessageCreating = (args) =>
            {
                return new CustomizedHttpWebRequestMessage(args,
                    response,
                    new Dictionary<string, string>()
                    {
                        {"Content-Type", "application/json;odata.metadata=minimal;odata.streaming=true;IEEE754Compatible=false;charset=utf-8"},
                        {"Preference-Applied", "odata.include-annotations=\"*\""}
                    });
            };
        }
    }
}